'''dossier.label.tests

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function
import time

from pyquchk import qc

from dossier.label import Label
from dossier.label.tests import coref_value, id_


@qc
def test_label_reverse_equality(cid1=id_, cid2=id_, ann=id_, v=coref_value):
    lab = Label(cid1, cid2, ann, v)
    assert lab == lab.reversed()


@qc
def test_label_hash_equality(cid1=id_, cid2=id_, ann=id_,
                             v1=coref_value, v2=coref_value):
    lab1, lab2 = Label(cid1, cid2, ann, v1), Label(cid1, cid2, ann, v2)
    assert lab1 == lab2 and hash(lab1) == hash(lab2)


@qc
def test_label_hash_equality_unordered(cid1=id_, cid2=id_, ann=id_,
                                       v1=coref_value, v2=coref_value):
    lab1, lab2 = Label(cid1, cid2, ann, v1), Label(cid2, cid1, ann, v2)
    assert lab1 == lab2 and hash(lab1) == hash(lab2)


@qc
def test_label_most_recent_first(cid1=id_, cid2=id_, ann=id_,
                                 v1=coref_value, v2=coref_value):
    t = int(time.time() * 1000)
    lab1 = Label(cid1, cid2, ann, v1, epoch_ticks=t)
    lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=t + 1)
    assert lab2 < lab1


@qc
def test_label_most_recent_first_unordered(cid1=id_, cid2=id_, ann=id_,
                                           v1=coref_value, v2=coref_value):
    t = int(time.time() * 1000)
    lab1 = Label(cid1, cid2, ann, v1, epoch_ticks=t)
    lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=t + 1)
    assert lab2 < lab1



'''dossier.label.label

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from collections import namedtuple
from datetime import datetime
import functools
from itertools import groupby, imap, combinations, ifilter
import logging
import sys
import time

import enum


logger = logging.getLogger(__name__)


MAX_SECOND_TICKS = ((60 * 60 * 24) * 365 * 100)
'''The maximum number of seconds supported.

Our kvlayer backend cannot (currently) guarantee a correct ordering
of signed integers, but can guarantee a correct ordering of unsigned
integers.

Labels, however, should be sorted with the most recent label first.
This is trivially possible by negating its epoch ticks.

Because kvlayer cannot guarantee a correct ordering of signed integers,
we avoid the sign switch by subtracting the ticks from an arbitrary
date in the future (UNIX epoch + 100 years).
'''


def time_complement(t):
    return MAX_SECOND_TICKS - t



class CorefValue(enum.Enum):
    '''
    An enumeration that describes the value of a coreference judgment by a
    human. The judgment is always made with respect to a pair of content
    items.

    :cvar Negative: The two items are not coreferent.
    :cvar Unknown: It is unknown whether the two items are coreferent.
    :cvar Positive: The two items are coreferent.
    '''
    Negative = -1
    Unknown = 0
    Positive = 1


@functools.total_ordering
class Label(namedtuple('_Label', 'content_id1 content_id2 subtopic_id1 ' +
                                 'subtopic_id2 annotator_id epoch_ticks '+
                                 'value')):
    '''A label is an immutable unit of ground truth data.

    .. automethod:: __new__
    .. automethod:: reversed
    .. automethod:: other
    .. automethod:: __lt__
    .. automethod:: __eq__
    .. automethod:: __hash__
    .. automethod:: __contains__
    '''

    def __new__(cls, content_id1, content_id2, annotator_id, value,
                subtopic_id1=None, subtopic_id2=None, epoch_ticks=None):
        # `__new__` is overridden instead of `__init__` because `namedtuple`
        # defines a `__new__` method. We modify construction by making
        # several values optional.
        if isinstance(value, int):
            value = CorefValue(value)
        if epoch_ticks is None:
            epoch_ticks = long(time.time())
        if subtopic_id1 is None:
            subtopic_id1 = ''
        if subtopic_id2 is None:
            subtopic_id2 = ''
        return super(Label, cls).__new__(
            cls, content_id1=content_id1, content_id2=content_id2,
            subtopic_id1=subtopic_id1, subtopic_id2=subtopic_id2,
            annotator_id=annotator_id, epoch_ticks=epoch_ticks, value=value)

    def reversed(self):
        '''Returns a new label with ids swapped.

        This method satisfies the following law: for every label
        ``lab``, ``lab == lab.reversed()``.
        '''
        return self._replace(
            content_id1=self.content_id2, content_id2=self.content_id1,
            subtopic_id1=self.subtopic_id2, subtopic_id2=self.subtopic_id1)

    def __contains__(self, content_id):
        '''Returns ``True`` if ``content_id`` is in this label.'''
        return content_id in (self.content_id2, self.content_id2)

    def other(self, content_id):
        '''Returns the other content id.

        If ``content_id == self.content_id1``, then return
        ``self.content_id2`` (and vice versa).
        '''
        if content_id == self.content_id1:
            return self.content_id2
        else:
            return self.content_id1

    def _to_kvlayer(self):
        '''Converts this label to a kvlayer tuple.

        The tuple returned can be used directly in a :mod:`kvlayer`
        ``put`` call.

        :rtype: ``(key, value)``
        '''
        epoch_ticks_rev = time_complement(self.epoch_ticks)
        negated = self._replace(epoch_ticks=epoch_ticks_rev)[0:len(self)-1]
        return (negated, str(self.value.value))

    @staticmethod
    def _from_kvlayer(row):
        '''Create a new :class:`Label` from a kvlayer result.

        The ``row`` should be a tuple of ``(key, value)``
        where ``key`` corresponds to the namespace defined at
        :attr:`LabelStore._kvlayer_namespace`.

        :param row: kvlayer result
        :type row: ``(key, value)``
        :rtype: :class:`Label`
        '''
        key, value = row
        cid1, cid2, subid1, subid2, ann, ticks = key
        return Label(content_id1=cid1, content_id2=cid2,
                     subtopic_id1=subid1, subtopic_id2=subid2,
                     annotator_id=ann,
                     epoch_ticks=time_complement(int(ticks)),
                     value=int(value))

    def __lt__(l1, l2):
        '''Defines a total ordering for labels.

        The ordering is meant to be the same as the ordering used
        in the underlying database storage. Namely, the key used
        to determine ordering is: ``(cid1, cid2, subid1, subid2,
        annotator_id, MAX_TIME - epoch_ticks)`` where ``cid1 <= cid2``
        and ``subid1 <= subid2``.

        Notably, the ordering does not include the coreferent ``value``
        and it complements ``epoch_ticks`` so that more recent
        labels appear first in ascending order.
        '''
        return l1._cmp_value < l2._cmp_value

    @property
    def _cmp_value(self):
        cid1, cid2 = normalize_pair(self.content_id1, self.content_id2)
        subid1, subid2 = normalize_pair(self.subtopic_id1, self.subtopic_id2)
        ticks = time_complement(self.epoch_ticks)
        return (cid1, cid2, subid1, subid2, self.annotator_id, ticks)

    def __eq__(l1, l2):
        '''Tests equality between two labels.

        Equality is keyed on ``annotator_id`` and the unordered
        comparison between content ids and subtopic ids.

        This definition of equality does not include the values for
        ``epoch_ticks`` or ``value``.
        '''
        return (
            l1.annotator_id == l2.annotator_id
            and unordered_pair_eq((l1.content_id1, l1.content_id2),
                                  (l2.content_id1, l2.content_id2))
            and unordered_pair_eq((l1.subtopic_id1, l1.subtopic_id2),
                                  (l2.subtopic_id1, l2.subtopic_id2))
        )

    def __hash__(self):
        '''Returns a hash of this label.

        The hash is made up of the content ids, subtopic ids and the
        annotator id. This hash function obeys the following law:
        for all labels ``x`` and ``y``, ``x == y`` if and only if
        ``hash(x) == hash(y)``.
        '''
        # This code is clearer if we use `frozenset`, but let's avoid
        # creating intermediate objects.
        cid1, cid2 = normalize_pair(self.content_id1, self.content_id2)
        subid1, subid2 = normalize_pair(self.subtopic_id1, self.subtopic_id2)
        return hash((self.annotator_id, cid1, cid2, subid1, subid2))

    def __repr__(self):
        tpl = 'Label({cid1}, {cid2}, ' + \
                    '{subid1}{subid2}annotator={ann}, {tstr}, value={v})'
        subid1, subid2 = '', ''
        if self.subtopic_id1:
            subid1 = 'subtopic1=%s, ' % self.subtopic_id1
        if self.subtopic_id2:
            subid2 = 'subtopic2=%s, ' % self.subtopic_id2
        dt = datetime.utcfromtimestamp(self.epoch_ticks)
        return tpl.format(cid1=self.content_id1, cid2=self.content_id2,
                         subid1=subid1, subid2=subid2, tstr=str(dt),
                         ann=self.annotator_id, v=self.value)


class LabelStore(object):
    '''A label database.

    .. automethod:: __init__
    .. automethod:: put
    .. automethod:: get
    .. automethod:: get_all_for_content_id
    .. automethod:: connected_component
    .. automethod:: expand
    .. automethod:: everything
    .. automethod:: delete_all
    '''
    config_name = 'dossier.label'
    TABLE = 'label'

    _kvlayer_namespace = {
        # (cid1, cid2, subid1, subid2, annotator_id, time) -> value
        # N.B. The `long` type here is for the benefit of the underlying
        # database storage. It will hopefully result in a storage type
        # that is big enough to contain milliseconds epoch ticks. (A 64 bit
        # integer is more than sufficient, which makes `long` unnecessary from
        # a Python 2 perspective.)
        TABLE: (str, str, str, str, str, long),
    }

    def __init__(self, kvlclient):
        '''Create a new label store.

        :param kvlclient: kvlayer client
        :type kvlclient: :class:`kvlayer._abstract_storage.AbstractStorage`
        :rtype: :class:`LabelStore`
        '''
        self.kvl = kvlclient
        self.kvl.setup_namespace(self._kvlayer_namespace)

    def put(self, label):
        '''Add a new label to the store.

        :param label: label
        :type label: :class:`Label`
        '''
        logger.info('adding label "%r"', label)
        self.kvl.put(self.TABLE,
                     label._to_kvlayer(), label.reversed()._to_kvlayer())

    def get(self, cid1, cid2, annotator_id, subid1='', subid2=''):
        '''Retrieve a label from the store.

        When ``subid1`` and ``subid2`` are empty, then a label without
        subtopic identifiers will be returned.

        Note that the combination of content ids, subtopic ids and
        an annotator id *uniquely* identifies a label.

        :param str cid1: content id
        :param str cid2: content id
        :param str annotator_id: annotator id
        :param str subid1: subtopic id
        :param str subid2: subtopic id
        :rtype: :class:`Label`
        :raises: :exc:`KeyError` if no label could be found.
        '''
        s = (cid1, cid2, subid1, subid2, annotator_id, long(-sys.maxint - 1))
        e = (cid1, cid2, subid1, subid2, annotator_id, long(sys.maxint))

        # We return the first result because the `kvlayer` abstraction
        # guarantees that the first result will be the most recent entry
        # for this particular key (since the timestamp is inserted as a
        # complement value).
        for row in self.kvl.scan(self.TABLE, (s, e)):
            return Label._from_kvlayer(row)
        raise KeyError((s, e))

    def get_all_for_content_id(self, content_id):
        '''Return a generator of labels connected to ``content_id``.

        If no labels are defined for ``content_id``, then the generator
        will yield no labels.

        Note that this only returns *directly* connected labels. It
        will not follow transitive relationships.

        :param str content_id: content id
        :rtype: generator of :class:`Label`
        '''
        s = (content_id,)
        e = (content_id + '\xff',)
        results = imap(Label._from_kvlayer, self.kvl.scan(self.TABLE, (s, e)))
        return latest_labels(results)

    def connected_component(self, content_id):
        '''Return a connected component generator for ``content_id``.

        Given a ``content_id`` and a coreference ``value`` of ``-1``,
        ``0`` or ``1``, return the corresponding connected component
        by following all transitivity relationships.

        For example, if ``(a, b, 1)`` is a label and ``(b, c, 1)`` is
        a label, then ``connected_component('a', 1)`` will return both
        labels even though ``a`` and ``c`` are not directly connected.

        The ``value`` indicates which labels to include in the
        connected component.

        (Note that even though this returns a generator, it will still
        consume memory proportional to the number of labels in the
        connected component.)

        :param str content_id: content id
        :param value: coreferent value
        :type value: :class:`CorefValue`
        :rtype: generator of :class:`Label`

        '''
        done = set()  # set of cids that we've queried with
        todo = set([content_id])  # set of cids to do a query for
        label_hashes = set()
        while todo:
            cid = todo.pop()
            done.add(cid)
            for label in self.get_all_for_content_id(cid):
                if label.value != CorefValue.Positive:
                    continue
                if label.content_id1 not in done:
                    todo.add(label.content_id1)
                if label.content_id2 not in done:
                    todo.add(label.content_id2)

                h = hash(label)
                if h not in label_hashes:
                    label_hashes.add(h)
                    yield label

    def expand(self, content_id):
        '''Return expanded set of labels from a connected component.

        The connected component is derived from ``content_id``.

        The labels returned by :meth:`LabelStore.connected_component`
        contains only the :class:`Label` stored in the
        :class:`LabelStore`, and does not include the labels you can
        infer from the connected component. This method returns both
        the data-backed labels and the inferred labels.

        Subtopic assignments of the expanded labels will be empty. The
        ``annotator_id`` will be an arbitrary ``annotator_id`` within
        the connected component.

        :param str content_id: content id
        :param value: coreferent value
        :type value: :class:`CorefValue`
        :rtype: ``list`` of :class:`Label`
        '''
        labels = list(self.connected_component(content_id))
        labels.extend(expand_labels(labels))
        return labels

    def negative_inference(self, content_id):
        '''Return a generator of inferred negative label relationships
        centered on ``content_id``.

        Negative labels are inferred by getting all other content ids
        connected to ``content_id`` through a negative label, then
        running :meth:`LabelStore.negative_label_inference` on those
        labels. See :meth:`LabelStore.negative_label_inference` for
        more information.
        '''
        neg_labels = ifilter(lambda l: l.value == CorefValue.Negative,
                             self.get_all_for_content_id(content_id))
        for label in neg_labels:
            label_inf = self.negative_label_inference(label)
            for label in label_inf:
                yield label

    def negative_label_inference(self, label):
        '''Return a generator of inferred negative label relationships.

        Construct ad-hoc negative labels between ``label.content_id1``
        and the positive connected component of ``label.content_id2``,
        and ``label.content_id2`` to the connected component of
        ``label.content_id1``.

        Note this will allocate memory proportional to the size of the
        connected components of ``label.content_id1`` and
        ``label.content_id2``.
        '''
        assert label.value == CorefValue.Negative

        cid1_comp = self.connected_component(label.content_id1)
        cid2_comp = self.connected_component(label.content_id2)

        def get_non_query_cid(cid, label):
            if label.content_id1 == cid:
                return label.content_id2
            else:
                return label.content_id1

        yield label

        for comp_label in cid2_comp:
            cid = get_non_query_cid(label.content_id2, comp_label)
            yield Label(label.content_id1, cid, 'auto', CorefValue.Negative)

        for comp_label in cid1_comp:
            cid = get_non_query_cid(label.content_id1, comp_label)
            yield Label(label.content_id2, cid, 'auto', CorefValue.Negative)

    def everything(self, include_deleted=False):
        '''Returns a generator of all labels in the store.

        If ``include_deleted`` is ``True``, labels that have been
        deleted are also included.

        :rtype: generator of :class:`Label`
        '''
        results = imap(Label._from_kvlayer, self.kvl.scan(self.TABLE))
        return results if include_deleted else latest_labels(results)

    def delete_all(self):
        '''Deletes all labels in the store.'''
        self.kvl.clear_table(self.TABLE)


def unordered_pair_eq(pair1, pair2):
    '''Performs pairwise unordered equality.

    ``pair1`` == ``pair2`` if and only if
    ``frozenset(pair1)`` == ``frozenset(pair2)``.
    '''
    (x1, y1), (x2, y2) = pair1, pair2
    return (x1 == x2 and y1 == y2) or (x1 == y2 and y1 == x2)


def normalize_pair(x, y):
    '''Normalize a pair of values.

    Returns ``(y, x)`` if and only if ``y < x`` and ``(x, y)``
    otherwise.
    '''
    if y < x:
        return y, x
    else:
        return x, y


def latest_labels(label_iterable):
    '''Returns the most recent labels from a sorted iterable.'''
    for _, group in groupby(label_iterable):
        for lab in group:
            yield lab
            break


def expand_labels(labels):
    '''Expand a set of labels that define a connected component.

    ``labels`` must define a *positive* connected component: it is all
    of the edges that make up the *single* connected component in the
    :class:`LabelStore`. expand will ignore subtopic assignments, and
    annotator_id will be an arbitrary one selected from ``labels``.

    Note that this function only returns the expanded labels, which
    is guaranteed to be disjoint with the given ``labels``. This
    requirement implies that ``labels`` is held in memory to ensure
    that no duplicates are returned.

    :param labels: iterable of :class:`Label` for the connected component.
    :rtype: generator of expanded :class:`Label`s only
    '''
    labels = list(labels)
    assert all(lab.value == CorefValue.Positive for lab in labels)

    # Anything to expand?
    if len(labels) == 0:
        return

    annotator = labels[0].annotator_id

    data_backed_pairs = set()
    connected_component = set()
    for label in labels:
        data_backed_pairs.add(
            normalize_pair(label.content_id1, label.content_id2))
        connected_component.add(label.content_id1)
        connected_component.add(label.content_id2)

    # We do not want to rebuild the Labels we already have,
    # because they have true annotator_id and subtopic
    # fields that we may want to preserve.
    for cid1, cid2 in combinations(connected_component, 2):
        if normalize_pair(cid1, cid2) not in data_backed_pairs:
            yield Label(cid1, cid2, annotator, CorefValue.Positive)



'''dossier.label.tests

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from pyquchk import qc
import pytest

from dossier.label import Label, LabelStore
from dossier.label.tests import kvl, coref_value, id_


@pytest.yield_fixture
def label_store(kvl):
    lstore = LabelStore(kvl)
    yield lstore
    lstore.delete_all()


def test_put_get(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()

        lab = Label(cid1, cid2, ann, v)
        label_store.put(lab)
        got = label_store.get(cid1, cid2, ann)
        assert lab == got and lab.value == got.value
    _()


def test_put_get_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()

        lab = Label(cid1, cid2, ann, v)
        label_store.put(lab)
        got = label_store.get(cid2, cid1, ann)
        assert lab == got and lab.value == got.value
    _()


def test_put_get_recent(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)
        got = label_store.get(cid1, cid2, ann)
        assert lab1 == got and lab2 == got and lab2.value == got.value
    _()


def test_put_get_recent_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)
        got = label_store.get(cid1, cid2, ann)
        assert lab1 == got and lab2 == got and lab2.value == got.value
    _()


def test_direct_connect_recent(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)

        direct = list(label_store.get_all_for_content_id(cid1))
        assert direct == [lab2] and direct == [lab1]
        assert direct[0].value == lab2.value

        direct = list(label_store.get_all_for_content_id(cid2))
        assert direct == [lab2] and direct == [lab1]
        assert direct[0].value == lab2.value
    _()


def test_direct_connect_recent_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)

        direct = list(label_store.get_all_for_content_id(cid1))
        assert direct == [lab2] and direct == [lab1]
        assert direct[0].value == lab2.value

        direct = list(label_store.get_all_for_content_id(cid2))
        assert direct == [lab2] and direct == [lab1]
        assert direct[0].value == lab2.value
    _()


def test_direct_connect(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    direct = list(label_store.get_all_for_content_id('a'))
    assert direct == [ab, ac]


def test_direct_connect_unordered(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('c', 'a', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    direct = list(label_store.get_all_for_content_id('a'))
    assert direct == [ab, ac]


def test_connected_component_basic(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, ac, bc])


def test_connected_component_unordered(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('c', 'a', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, ac, bc])


def test_connected_component_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', -1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc])


def test_connected_component_many(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])


def test_connected_component_many_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', -1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])


def test_connected_component_many_most_recent(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', -1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])

    # This label should overwrite the existing `bc` label and expand
    # the connected component to `cd` through transitivity.
    bc = Label('b', 'c', '', 1, epoch_ticks=bc.epoch_ticks + 1)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])


def test_connected_component_many_most_recent_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])

    # This label should overwrite the existing `bc` label and contract
    # the connected component to just `ab`.
    bc = Label('b', 'c', '', -1, epoch_ticks=bc.epoch_ticks + 1)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])


def test_expand(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    ae = Label('a', 'e', '', -1)
    fg = Label('f', 'g', '', 1)

    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)
    label_store.put(ae)
    label_store.put(fg)

    def get_pair(label):
        return (label.content_id1, label.content_id2)

    correct_pairs = [('a', 'b'),
                     ('a', 'c'),
                     ('a', 'd'),
                     ('b', 'c'),
                     ('b', 'd'),
                     ('c', 'd')]

    expansion = label_store.expand('a')

    assert frozenset(map(get_pair, expansion)) == \
        frozenset(correct_pairs)

    expansion = label_store.expand('e')

    assert frozenset(map(get_pair, expansion)) == frozenset([])

    expansion = label_store.expand('f')

    correct_pairs = [('f', 'g')]

    assert frozenset(map(get_pair, expansion)) == frozenset(correct_pairs)


def test_negative_label_inference(label_store):
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)

    de = Label('d', 'e', '', 1)
    df = Label('d', 'f', '', 1)
    dg = Label('d', 'g', '', -1)

    cd = Label('c', 'd', '', -1)

    label_store.put(ac)
    label_store.put(bc)
    label_store.put(de)
    label_store.put(df)
    label_store.put(cd)
    label_store.put(dg)

    def get_pair(label):
        return (label.content_id1, label.content_id2)

    correct_pairs = [('c', 'e'),
                     ('c', 'f'),
                     ('d', 'a'),
                     ('d', 'b'),
                     ('c', 'd')]

    inference = label_store.negative_label_inference(cd)

    assert frozenset(map(get_pair, inference)) == \
        frozenset(correct_pairs)


def test_negative_inference(label_store):
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)

    de = Label('d', 'e', '', 1)
    df = Label('d', 'f', '', 1)

    cg = Label('c', 'g', '', -1)
    dg = Label('d', 'g', '', -1)

    hg = Label('h', 'g', '', 1)

    label_store.put(ac)
    label_store.put(bc)
    label_store.put(de)
    label_store.put(df)
    label_store.put(cg)
    label_store.put(dg)
    label_store.put(hg)

    def get_pair(label):
        return (label.content_id1, label.content_id2)

    correct_pairs = [('g', 'a'),
                     ('g', 'b'),
                     ('c', 'h'),
                     ('g', 'e'),
                     ('g', 'f'),
                     ('d', 'h'),
                     ('g', 'c'),
                     ('g', 'd')]

    inference = label_store.negative_inference('g')

    assert frozenset(map(get_pair, inference)) == \
        frozenset(correct_pairs)



'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

``dossier.label`` command line tool
===================================

``dossier.label`` is a command line application for viewing the raw label
data inside the database. Generally, this is a debugging tool for developers.

Run ``dossier.label --help`` for the available commands.
'''
from __future__ import absolute_import, division, print_function
import argparse
from itertools import groupby, imap, islice
import json
import sys

import kvlayer
import yakonfig

from dossier.label import Label, LabelStore


def label_to_dict(lab):
    return {field: getattr(lab, field) for field in lab._fields}


def dict_to_label(d):
    def to_bytes(v):
        if isinstance(v, unicode):
            return v.encode('utf-8')
        return v
    
    def to_long(v):
        if isinstance(v, int):
            return long(v)
        return v

    return Label(**{k: to_long(to_bytes(v)) for k, v in d.items()})


class App(yakonfig.cmd.ArgParseCmd):
    def __init__(self, *args, **kwargs):
        yakonfig.cmd.ArgParseCmd.__init__(self, *args, **kwargs)
        self._label_store = None

    @property
    def label_store(self):
        if self._label_store is None:
            self._label_store = LabelStore(kvlayer.client())
        return self._label_store

    def args_list(self, p):
        p.add_argument('--include-deleted', action='store_true',
                       help='When set, show deleted labels.')

    def do_list(self, args):
        labels = self.label_store.everything(
            include_deleted=args.include_deleted)
        for k, group in groupby(sorted(labels)):
            print(k)
            for lab in islice(group, 1, None):
                print('    %s' % repr(lab))

    def args_dump_all(self, p):
        p.add_argument('--exclude-deleted', action='store_true',
                       help='When set, only the most recent labels are '
                            'dumped.')

    def do_dump_all(self, args):
        labels = list(imap(label_to_dict, self.label_store.everything(
            include_deleted=not args.exclude_deleted)))
        json.dump(labels, fp=sys.stdout)

    def args_load(self, p):
        p.add_argument('fpath', nargs='?', default=None,
                       help='File path containing label data. When absent, '
                            'stdin is used.')

    def do_load(self, args):
        fout = sys.stdin if args.fpath is None else open(args.fpath, 'w+')
        for lab in imap(dict_to_label, json.load(fp=fout)):
            self.label_store.put(lab)

    def args_get(self, p):
        p.add_argument('content_id', type=str,
                       help='Show all labels directly ascribed to content_id')
        p.add_argument('--value', type=CorefValue, choices=[-1, 0, 1],
                       default=None,
                       help='Only show labels with this coreferent value.')

    def do_get(self, args):
        for label in self.label_store.get_all_for_content_id(args.content_id):
            if args.value is None or label.value == args.value:
                print(label)

    def args_connected(self, p):
        p.add_argument('content_id', type=str,
                       help='Show all labels directly ascribed to content_id')
        p.add_argument('value', type=int, choices=[-1, 0, 1],
                       help='Only show labels with this coreferent value.')

    def do_connected(self, args):
        connected = self.label_store.connected_component(
            args.content_id, args.value)
        for label in connected:
            print(label)

    def args_delete_all(self, p):
        pass

    def do_delete_all(self, args):
        answer = raw_input('Are you absolutely sure? [y/n] ').strip().lower()
        if answer and answer[0] == 'y':
            self.label_store.delete_all()


def main():
    p = argparse.ArgumentParser(
        description='Interact with DossierStack truth data.')
    app = App()
    app.add_arguments(p)
    args = yakonfig.parse_args(p, [kvlayer, yakonfig])
    app.main(args)
